import pytest
import wx


@pytest.fixture()
def window():
    app = wx.App()
    frame = wx.Frame(None)
    yield frame
    wx.CallAfter(frame.Close)
    app.MainLoop()


@pytest.fixture()
def main_window():
    from pyguitemp.core import MainWindow

    app = wx.App()
    frame = MainWindow(None, "My App")
    yield frame
    wx.CallAfter(frame.Close)
    app.MainLoop()


@pytest.fixture()
def plugin():
    from typing import List

    import wx

    from pyguitemp.plugins import MenuTool, PluginBase

    class ExamplePlugin(PluginBase):
        def central(self, parent=None) -> wx.Window:
            return wx.TextCtrl(parent, style=wx.TE_MULTILINE)

        def toolbar_items(self) -> List[MenuTool]:
            return [
                MenuTool(
                    id=wx.ID_STOP,
                    text="Exit",
                    short_help="Terminate application",
                    bitmap=wx.ArtProvider.GetBitmap(
                        wx.ART_QUIT, wx.ART_TOOLBAR, wx.Size(50, 50)
                    ),
                ),
            ]

    return ExamplePlugin


@pytest.fixture()
def empty_plugin():
    from pyguitemp.plugins import PluginBase

    class EmptyPlugin(PluginBase):
        pass

    return EmptyPlugin



from unittest.mock import MagicMock, patch


class TestPluginBase:
    def test_plugin_registration(self, plugin):
        from pyguitemp.plugins import KNOWN_PLUGINS

        assert plugin in KNOWN_PLUGINS

    def test_menu_entries(self, empty_plugin):
        """Dummy tests to check default outputs"""
        assert empty_plugin().menu_entries() == []

    def test_toolbar_items(self, empty_plugin):
        """Dummy tests to check default outputs"""
        assert empty_plugin().toolbar_items() == []

    def test_tabs(self, empty_plugin):
        """Dummy tests to check default outputs"""
        assert empty_plugin().tabs() == []

    def test_central(self, empty_plugin):
        """Dummy tests to check default outputs"""
        assert empty_plugin().central() is None


def test_collect_plugins():
    from pathlib import Path

    from pyguitemp.plugins import collect_plugins

    this_file = Path(__file__)
    plugins = collect_plugins(this_file.parent)
    assert f"{this_file.parent.stem}.{this_file.stem}" in plugins

    plugins = collect_plugins(this_file.parent, package="some_package")
    assert f"some_package.{this_file.parent.stem}.{this_file.stem}" in plugins

    plugins = collect_plugins(this_file.parent, add_to_path=True)
    assert f"{this_file.stem}" in plugins


def test_collect_builtin_extensions():
    with patch("pyguitemp.plugins.collect_plugins", MagicMock()):
        from pyguitemp.plugins import collect_builtin_extensions, collect_plugins

        collect_builtin_extensions()
        collect_plugins.assert_called_once()


def test_load_plugins(caplog):
    with patch("importlib.import_module", MagicMock()):
        from importlib import import_module

        from pyguitemp.plugins import collect_builtin_extensions, load_plugins

        plugins = collect_builtin_extensions()
        load_plugins(plugins)
        import_module.assert_called()

    plugins.append("Wrong plugin")
    load_plugins(plugins)
    assert "Plugin 'Wrong plugin' could not be loaded." in caplog.messages[-1]



"""Adapted from ZetCode wxPython tutorial

https://zetcode.com/wxpython/dialogs/

In this example, we create an
about dialog box.

author: Jan Bodnar
website: www.zetcode.com
last modified: July 2020
"""
from pathlib import Path
from typing import List

import wx
import wx.adv

from ...plugins import MenuTool, PluginBase


class AboutDialogPlugin(PluginBase):
    def menu_entries(self) -> List[MenuTool]:
        return [
            MenuTool(
                menu="About",
                id=wx.ID_ABOUT,
                text="About",
                description="Some info about this app",
                callback=self.OnAboutBox,
            ),
        ]

    def OnAboutBox(self, e):
        with (Path(__file__).parent / "description").open("r") as f:
            description = f.read()

        with (Path(__file__).parent / "licence").open("r") as f:
            licence = f.read()

        info = wx.adv.AboutDialogInfo()

        info.SetIcon(
            wx.Icon(str(Path(__file__).parent / "logo.png"), wx.BITMAP_TYPE_PNG)
        )
        info.SetName("File Hunter")
        info.SetVersion("1.0")
        info.SetDescription(description)
        info.SetCopyright("(C) 2007 - 2021 Jan Bodnar")
        info.SetWebSite("http://www.zetcode.com")
        info.SetLicence(licence)
        info.AddDeveloper("Jan Bodnar")
        info.AddDocWriter("Jan Bodnar")
        info.AddArtist("The Tango crew")
        info.AddTranslator("Jan Bodnar")

        wx.adv.AboutBox(info)



"""
The entry point of the program. It launches the main application - i.e. it starts
whatever represents the mainloop in the chosen GUI toolkit.
"""
from .scripts import main

if __name__ == "__main__":
    main()



"""
Contains the definition of the top window of the program, including the menu bar, the
toolbar (if any), the central widget, bottom progress bar and status area.

In particular, the progress bar and the status area are the kind of things that the
plugins will be importing - or at least wanting to update - regularly. They are defined
as global aobject singletons such that even though an instance is created once, such an
instance can be access directly from the class anywhere else where you import the class.
"""
from __future__ import annotations

import itertools
from typing import Dict, List, Optional

import wx

from .logging import logger
from .plugins import KNOWN_PLUGINS, MenuTool, PluginBase, load_plugins


class StatusBar(wx.StatusBar):
    """
    Singleton class to manage the top level status and progress bar.

    It provides two fields:
        - field 0: Automatic display of menu and tools descriptions as well as custom
            status messages
        - field 1: Holds the progress bar. Do not use it for displaying text.

    Use 'StatusBar().SomeMethod` to manage the status bar messages and properties. See
    'wx.StatusBar' for more information about the options available.

    Use `StatusBar().progress_bar.SomeMethod` to manage the progress bar properties.
    See 'wx.Gauge' for more information about the options available.
    """

    _instance: Optional[StatusBar] = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = wx.StatusBar.__new__(cls)
        return cls._instance

    def __init__(self, *args, progress_bar_width: int = 150, **kwargs):
        super(StatusBar, self).__init__(*args, **kwargs)
        self.SetFieldsCount(2, (-1, progress_bar_width))
        self.progress_bar = wx.Gauge(
            self,
            wx.ID_ANY,
            size=(progress_bar_width, 10),
            style=wx.GA_HORIZONTAL | wx.GA_SMOOTH,
        )
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        sizer.Add(wx.StaticText(self, -1, ""), 1, wx.ALL)
        sizer.Add(self.progress_bar, 0, wx.ALL, 10)
        self.SetSizer(sizer)
        self.Layout()

    def SetStatusWidths(self, width: int):
        """Ensures that progress bar is resized if fields are resized.

        Args:
            width: Width of the progress bar.
        """
        super(StatusBar, self).SetStatusWidths((-1, width))
        self.progress_bar.SetSize(self.GetStatusWidth(1), 10)


class MainWindow(wx.Frame):
    def __init__(
        self,
        parent,
        title: str,
        notebook_layout: bool = True,
        tab_style: int = wx.NB_TOP,
    ):
        super(MainWindow, self).__init__(parent, title=title)
        self.notebook_layout = notebook_layout
        self.tab_style = tab_style
        self.SetStatusBar(StatusBar(self))

    def populate_window(self):
        """Adds menu items, tools and other widgets in plugins to the main window."""
        if self.notebook_layout:
            self._make_notebook(self.tab_style)
        else:
            self._make_central_widget()
        self._make_toolbar()
        self._make_menubar()
        self.SetInitialSize(wx.Size(800, 600))

    def _make_menubar(self) -> None:
        """Create the menu bar from the entries provided by the widgets."""
        # Collecting the menu entries
        entries = itertools.chain.from_iterable(
            [view().menu_entries() for view in KNOWN_PLUGINS]
        )

        # Creating the menus
        menus = dict()
        for entry in entries:
            if entry.menu not in menus:
                menus[entry.menu] = wx.Menu()

            menu_entry = menus[entry.menu].Append(
                entry.id, entry.text, entry.description, entry.kind
            )

            if entry.callback is not None:
                self.Bind(wx.EVT_MENU, entry.callback, menu_entry)

        # Creating the menubar.
        menu_bar = wx.MenuBar()
        for name, item in menus.items():
            menu_bar.Append(item, name)

        # Adding the MenuBar to the Frame content.
        self.SetMenuBar(menu_bar)

    def _make_toolbar(self):
        """Create the tool bar from the entries provided by the widgets."""
        # Collect tools
        tools = itertools.chain.from_iterable(
            [view().toolbar_items() for view in KNOWN_PLUGINS]
        )

        # Including the tools
        toolbar = self.CreateToolBar()
        for tool in tools:
            item = toolbar.AddTool(
                tool.id, tool.text, tool.bitmap, tool.short_help, tool.kind
            )

            if tool.callback is not None:
                self.Bind(wx.EVT_MENU, tool.callback, item)

        toolbar.Realize()

    def _make_notebook(self, tab_style: int = wx.NB_TOP) -> None:
        """Create the central widget of the window as a notebook.

        A notebook is created as the central widget and any other view provided by the
        plugins is added as new page. Finally, the first page is selected.

        Args:
            tab_style: integer indicating the position of the tabs. Valid values (OS
                dependent) are wx.NB_TOP, wx.NB_LEFT, wx.NB_RIGHT, wx.NB_BOTTOM,
                wx.NB_FIXEDWIDTH, wx.NB_MULTILINE and wx.NB_NOPAGETHEME.
        """
        self.notebook = wx.Notebook(self, style=tab_style)

        # Collect tabs
        tabs = itertools.chain.from_iterable(
            [view().tabs(self.notebook) for view in KNOWN_PLUGINS]
        )

        # Add tabs to notebook
        for tab in sorted(tabs, key=lambda x: x.order):
            self.notebook.AddPage(tab.page, tab.text, tab.select, tab.imageId)

        if self.notebook.PageCount > 0:
            self.notebook.SetSelection(0)

    def _make_central_widget(self) -> None:
        """Create the central widget of the window.

        Such a central widget is taken from the plugins. Exactly one central widget
        needs to be provided between all plugins.

        Raises:
            ValueError: If the number of central widgets found is not 1.
        """
        widget = [
            v for v in [view().central(self) for view in KNOWN_PLUGINS] if v is not None
        ]

        if (n := len(widget)) != 1:
            raise ValueError(
                f"Exactly 1 central widget needs to be provided. {n} given."
            )


class BuiltInActions(PluginBase):
    """Actions and features builtin with the core window."""

    def menu_entries(self) -> List[MenuTool]:
        return [
            MenuTool(
                menu="File",
                id=wx.ID_EXIT,
                text="Exit",
                description="Terminate application",
            ),
        ]


_tab_location: Dict[str, int] = {
    "top": wx.NB_TOP,
    "bottom": wx.NB_BOTTOM,
    "left": wx.NB_LEFT,
    "right": wx.NB_RIGHT,
}


class MainApp(wx.App):
    def __init__(
        self,
        *args,
        title: str,
        plugins_list: Optional[List[str]] = None,
        notebook_layout: bool = True,
        tab_style: str = "top",
        **kwargs,
    ):
        self.title = title
        self.plugins_list = plugins_list if plugins_list is not None else []
        self.notebook_layout = notebook_layout
        try:
            self.tab_style = _tab_location[tab_style]
        except KeyError:
            logger.warning(
                f"Invalid tab_style '{tab_style}'. Valid values are "
                f"{list(_tab_location.keys())}. Defaulting to 'top'"
            )
            self.tab_style = _tab_location["top"]

        super(MainApp, self).__init__(*args, **kwargs)

    def OnInit(self) -> bool:
        self.SetAppName(self.title)
        window = MainWindow(None, self.title, self.notebook_layout, self.tab_style)
        self.SetTopWindow(window)

        load_plugins(self.plugins_list)
        window.populate_window()
        window.Show(True)

        return True



from unittest.mock import MagicMock, patch

import pytest


class TestStatusBar:
    def test_set_status_widths(self, window):
        import wx

        from pyguitemp.core import StatusBar

        bar = StatusBar(window)
        assert isinstance(bar.progress_bar, wx.Gauge)

        expected = 300
        bar.SetStatusWidths(expected)
        assert bar.GetStatusWidth(1) == expected
        assert bar.progress_bar.GetSize()[0] == expected


class TestMainWindow:
    def test_populate_window(self, main_window):
        main_window._make_notebook = MagicMock()
        main_window._make_toolbar = MagicMock()
        main_window._make_menubar = MagicMock()
        main_window._make_central_widget = MagicMock()

        main_window.populate_window()
        main_window._make_menubar.assert_called_once()
        main_window._make_toolbar.assert_called_once()
        main_window._make_notebook.assert_called_once()
        main_window._make_central_widget.assert_not_called()

        main_window.notebook_layout = False
        main_window.populate_window()
        main_window._make_central_widget.assert_called_once()

    def test__make_menubar(self, main_window):
        main_window._make_menubar()
        menus = main_window.GetMenuBar().GetMenus()
        assert len(menus) > 0
        assert menus[0][1] == "File"

    def test__make_toolbar(self, main_window):
        main_window._make_toolbar()
        assert main_window.GetToolBar().GetToolsCount() == 0

    def test__make_notebook(self, main_window):
        import wx

        main_window._make_notebook()
        assert isinstance(main_window.notebook, wx.Notebook)
        assert main_window.notebook.PageCount == 0

    def test__make_central_widget(self, main_window):
        import wx

        from pyguitemp.plugins import PluginBase

        with pytest.raises(ValueError):
            main_window._make_central_widget()

        class SomePlugin(PluginBase):
            def central(self, parent):
                return wx.TextCtrl(parent, style=wx.TE_MULTILINE)

        main_window._make_central_widget()

        class SomeOtherPlugin(PluginBase):
            def central(self, parent):
                return wx.TextCtrl(parent, style=wx.TE_MULTILINE)

        with pytest.raises(ValueError):
            main_window._make_central_widget()


class TestBuiltInActions:
    def test_menu_entries(self):
        from pyguitemp.core import BuiltInActions
        from pyguitemp.plugins import MenuTool

        entries = BuiltInActions().menu_entries()
        assert len(entries) > 0
        assert all([isinstance(item, MenuTool) for item in entries])
        assert entries[0].menu == "File"


class TestMainApp:
    def test_on_init(self, caplog):
        import wx

        with patch("pyguitemp.core.MainWindow.Show", MagicMock()):
            from pyguitemp.core import MainApp, MainWindow

            app = MainApp(title="Some App", tab_style="a corner")
            assert "Invalid tab_style" in caplog.messages[-1]
            assert app.tab_style == wx.NB_TOP
            assert isinstance(app.GetTopWindow(), MainWindow)

            wx.CallAfter(app.GetTopWindow().Close)
            app.MainLoop()



from unittest.mock import MagicMock, patch

from pytest import mark


def test_thread_result():
    from pyguitemp.threads import ThreadResult

    result = ThreadResult("some data", 1)
    assert result.data == "some data"
    assert result.GetEventType() == 1


class TestWorkerThread:
    def test_connect_events(self, window):
        from pyguitemp.threads import WorkerThread

        window.Connect = MagicMock()
        worker = WorkerThread(lambda: None)

        worker.connect_events(window)
        assert window.Connect.call_count == 3

    def test_run(self):
        class Pool:
            post_event = MagicMock()

        with patch("pyguitemp.threads.ThreadResult", MagicMock()), patch(
            "pyguitemp.threads.ThreadPool", Pool
        ):
            from pyguitemp.threads import ThreadResult, WorkerThread

            result = "some result"
            worker = WorkerThread(lambda: result)

            worker.run()
            ThreadResult.assert_called_once_with(result, worker._event_on_complete)
            ThreadResult.reset_mock()

            worker.abort = True
            worker.run()
            ThreadResult.assert_called_once_with(result, worker._event_on_abort)
            ThreadResult.reset_mock()

            def error():
                raise ValueError("Error msg")

            worker = WorkerThread(error)
            worker.run()
            ThreadResult.assert_called_once_with("Error msg", worker._event_on_error)

    @mark.parametrize("callback", ["on_abort", "on_complete", "on_error"])
    def test_callbacks(self, callback):
        from pyguitemp.threads import WorkerThread

        class Event:

            _data = MagicMock()

            @property
            def data(self):
                return self._data()

        event = Event()
        worker = WorkerThread(lambda: None)
        getattr(worker, callback)(event)
        event._data.assert_not_called()

        worker = WorkerThread(lambda: None, **{callback: MagicMock()})
        getattr(worker, callback)(event)
        event._data.assert_called_once()


class TestThreadPool:
    def test_run_thread(self, window):
        class Worker:
            ident = 42
            abort = False
            connect_events = MagicMock()
            start = MagicMock()

        with patch("pyguitemp.threads.WorkerThread", MagicMock(return_value=Worker)):
            from pyguitemp.threads import ThreadPool

            pool = ThreadPool(window)
            pool.run_thread(lambda: None)

        assert Worker.ident in pool._workers

    def test_query_abort(self, window):
        import threading

        class Worker:
            ident = threading.get_ident()
            abort = False
            connect_events = MagicMock()
            start = MagicMock()

        with patch("pyguitemp.threads.WorkerThread", MagicMock(return_value=Worker)):
            from pyguitemp.threads import ThreadPool

            pool = ThreadPool(window)
            pool.run_thread(lambda: None)

        assert not pool.query_abort()

    def test_abort_thread(self, window):
        import threading

        class Worker:
            ident = threading.get_ident()
            abort = False
            connect_events = MagicMock()
            start = MagicMock()

        with patch("pyguitemp.threads.WorkerThread", MagicMock(return_value=Worker)):
            from pyguitemp.threads import ThreadPool

            pool = ThreadPool(window)
            ident = pool.run_thread(lambda: None)

        pool.abort_thread(ident)
        assert pool._workers[ident].abort

    def test_post_event(self, window):
        class WX:
            PostEvent = MagicMock()

        with patch("pyguitemp.threads.wx", WX):
            from pyguitemp.threads import ThreadPool

            pool = ThreadPool(window)
            pool.post_event(None)
            WX.PostEvent.assert_called_once()



"""
Contains all the machinery for dealing with threads.
"""
from __future__ import annotations

import threading
from typing import Any, Callable, Dict, Optional

import wx


class ThreadResult(wx.PyEvent):
    """Simple event to carry arbitrary result data."""

    def __init__(self, data: Any, event_type: int):
        """Initialises a result event for the thread.

        Args:
            data: The data to be passed to the main thread.
            event_type: The type of event.
        """
        wx.PyEvent.__init__(self)
        self.SetEventType(event_type)
        self.data = data


class WorkerThread(threading.Thread):
    """Worker Thread Class."""

    def __init__(
        self,
        target: Callable,
        on_abort: Optional[Callable] = None,
        on_complete: Optional[Callable] = None,
        on_error: Optional[Callable] = None,
        daemon: Optional[bool] = None,
    ):
        """Configures and start a new thread executing the target action.

        Args:
            target: The function to be executed
            on_abort: Function to be called if the thread aborts. Takes as only argument
                the output of the target function when that one aborts.
            on_complete: Function to be called if the thread finished normally. Takes as
                only argument the output of the target function when that one finishes.
            on_error: Function to be called if anything bad happens in the thread. Takes
                as only argument the exception catched.
        """
        super(WorkerThread, self).__init__(target=target, daemon=daemon)
        self._on_abort = on_abort
        self._on_complete = on_complete
        self._on_error = on_error
        self._event_on_abort = wx.NewEventType()
        self._event_on_complete = wx.NewEventType()
        self._event_on_error = wx.NewEventType()
        self.abort = False

    def connect_events(self, window: wx.Frame) -> None:
        """Connect the complete, abort and error events to the functions to execute.

        Args:
            window: The main window of the program.
        """
        window.Connect(-1, -1, self._event_on_abort, self.on_abort)
        window.Connect(-1, -1, self._event_on_complete, self.on_complete)
        window.Connect(-1, -1, self._event_on_error, self.on_error)

    def run(self):
        """Run worker thread and deals with the wrapping up accordingly."""
        try:
            result = self._target()

            if self.abort:
                ThreadPool().post_event(ThreadResult(result, self._event_on_abort))
            else:
                ThreadPool().post_event(ThreadResult(result, self._event_on_complete))

        except Exception as err:
            ThreadPool().post_event(ThreadResult(err.args[0], self._event_on_error))

    def on_abort(self, event: ThreadResult):
        """To execute when the thread execution is aborted.

        Args:
            event: The event object with the relevant output data
        """
        if self._on_abort is not None:
            self._on_abort(event.data)

    def on_complete(self, event: ThreadResult):
        """To execute when the thread execution is completed normally.

        Args:
            event: The event object with the relevant output data
        """
        if self._on_complete is not None:
            self._on_complete(event.data)

    def on_error(self, event: ThreadResult):
        """To execute when the thread ends with an error.

        Args:
            event: The exception raised
        """
        if self._on_error is not None:
            self._on_error(event.data)


class ThreadPool:

    _instance: Optional[ThreadPool] = None

    def __new__(cls, window: Optional[wx.Frame] = None):
        if window is None and cls._instance is None:
            raise ValueError(
                "The first time it is called, 'window' must be a wx.Frame object."
            )
        elif cls._instance is None:
            cls._instance = object.__new__(cls)
            cls._instance._window = window
            cls._instance._workers = {}
        return cls._instance

    def __init__(self, window: Optional[wx.Frame] = None):
        self._window: wx.Frame
        self._workers: Dict[int, WorkerThread]

    def run_thread(
        self,
        target: Callable,
        on_abort: Optional[Callable] = None,
        on_complete: Optional[Callable] = None,
        on_error: Optional[Callable] = None,
        daemon: Optional[bool] = None,
    ) -> int:
        """Runs a new thread executing the target callable on it.

        The thread will run until:
            - The 'target' function returns
            - An exception is raised

        The target function might finish because it has completed what it was doing. In
        that case, 'on_complete' is executed after returning from `target`.

        It might also finish because it is monitoring the value of
        `ThreadPool().query_abort()` and this is `True`. In that case, target should
        return with an appropriate value and `on_abort` is then executed.

        Finally, an exception might occur in the thread. In this case, the exception
        is caught and `on_error` is executed.

        `on_complete`, `on_abort` and `on_error` are all executed in the main thread.

        Args:
            target: The function to be executed in a separate thread.
            on_abort: The function to be executed when the target function is aborted.
                Takes as input the value returned by target.
            on_complete: The function to be executed when the target function is
                completed normally. Takes as input the value returned by target.
            on_error: The function to be executed when an exception is raised in the
                target. Takes as input the exception raised.
            daemon: If the function is a daemonic function.

        Returns:
            The id number for the thread, needed if it is to be aborted externally.
        """
        thread = WorkerThread(target, on_abort, on_complete, on_error, daemon)
        thread.connect_events(self._window)
        thread.start()
        self._workers[thread.ident] = thread
        return thread.ident

    def query_abort(self) -> bool:
        """Checks if the current thread is to be aborted."""
        return self._workers[threading.get_ident()].abort

    def abort_thread(self, ident: int) -> None:
        """Flags the thread with `ident` to be aborted.

        Args:
            ident: The id of the thread
        """
        self._workers[ident].abort = True

    def post_event(self, event: ThreadResult):
        """Adds an event to the event loop of the main thread."""
        wx.PostEvent(self._window, event)



from __future__ import annotations

import argparse
from abc import ABC, abstractmethod
from pathlib import Path
from shutil import copytree
from typing import Dict, List

from . import APP_NAME
from .config import APP_LONG_NAME, AUTO_PLUGINS, NOTEBOOK_LAYOUT, PLUGINS, TAB_STYLE
from .core import MainApp
from .logging import logger

logger.app_name = APP_NAME


def _run_app():
    """Runs pyguitemp as an application, loading all the plugins available."""
    app = MainApp(
        title=APP_LONG_NAME,
        plugins_list=PLUGINS + AUTO_PLUGINS,
        notebook_layout=NOTEBOOK_LAYOUT,
        tab_style=TAB_STYLE,
    )
    app.MainLoop()


def _init_repo(path: Path, name: str):
    """Initialises repository with the files/folders required to use pyguitemp.

    This is done by copying the skeleton repository structure within pyguitemp to the
    desired location.

    Args:
        path: Where the repository should be initialized. Default: current.
        name: Name of the application.
    """
    full_dir = path / name
    logger.info(f"Initialising '{name}' in `{full_dir}`...")
    copytree(Path(__file__).parent / "skeleton", full_dir)


class SubCommand(ABC):
    def __init__(self, name, description):
        self.name = name
        self.description = description

    @abstractmethod
    def add_arguments(self, parser: argparse.ArgumentParser):
        pass

    @abstractmethod
    def run(self, args: argparse.Namespace):
        pass


class RunSubCommand(SubCommand):
    def __init__(self):
        super().__init__("run", f"Run '{APP_NAME}' as a standalone app.")

    def add_arguments(self, parser: argparse.ArgumentParser):
        pass

    def run(self, args: argparse.Namespace):
        _run_app()


class InitSubCommand(SubCommand):
    def __init__(self):
        super().__init__(
            "init", "Initialises the current directory with an skeleton of a GUI app"
        )

    def add_arguments(self, parser: argparse.ArgumentParser):
        parser.add_argument(
            "name",
            type=str,
            help="Name of the app. To be used as package name.",
        )
        parser.add_argument(
            "-t",
            "--target",
            default=".",
            type=str,
            help="Target location for the app. Default: Current directory.",
        )

    def run(self, args: argparse.Namespace):
        _init_repo(Path(args.target).absolute(), args.name)


_SUB_COMMANDS: List[SubCommand] = [RunSubCommand(), InitSubCommand()]

_SUB_COMMAND_BY_NAME: Dict[str, SubCommand] = {
    sub_command.name: sub_command for sub_command in _SUB_COMMANDS
}


def _parse_args(argv: List[str] = None) -> argparse.Namespace:
    """Parse the input arguments list populating the relevant subcommands

    Args:
        argv: List of arguments

    Returns:
        The namespace containing the relevant sub command information and arguments.
    """
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest="command")
    subparsers.required = True
    for sub_command in _SUB_COMMANDS:
        sub_parser = subparsers.add_parser(
            sub_command.name, help=sub_command.description
        )
        sub_command.add_arguments(sub_parser)

    return parser.parse_args(argv)


def main(argv: List[str] = None):
    """Main entry point for pyguitemp, calling the relevant subcommands.

    Args:
        argv: List of input arguments, including subcommand.

    Returns:

    """
    args = _parse_args(argv)
    _SUB_COMMAND_BY_NAME[args.command].run(args)



"""
The entry point of the program. It launches the main application - i.e. it starts
whatever represents the mainloop in the chosen GUI toolkit.
"""
from pyguitemp.core import MainApp

from .config import APP_LONG_NAME, AUTO_PLUGINS, NOTEBOOK_LAYOUT, PLUGINS, TAB_STYLE

app = MainApp(
    title=APP_LONG_NAME,
    plugins_list=PLUGINS + AUTO_PLUGINS,
    notebook_layout=NOTEBOOK_LAYOUT,
    tab_style=TAB_STYLE,
)
app.MainLoop()
